#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "StormLib.h"
#include "tinyxml.h"

#include "outstreams.hpp"
#include "coordinates.hpp"
#include "SC2Map.hpp"



int SC2Map::readMap( const char* strArchiveFilename )
{
  HANDLE archive;

  if( !SFileOpenArchive( strArchiveFilename, 0, 0, &archive ) )
  {
    printWarning( "Could not open map MPQ archive file %s\n", strArchiveFilename );
    return -1;
  }

  // keep the archive name for output, so users know
  sprintf( strArchiveName, "%s", strArchiveFilename );

  // the order of reading archive files is important--some
  // analysis data structures are depedenent on one another

  if( readGameStrings_txt( archive ) < 0 )
  {
    printWarning( "Could not extract map name from archive, using archive filename instead.\n" );
    setMapNameFromArchiveName( strArchiveFilename );
  }
  // however we got a name for the map, clean it up
  // as a no spaces, char first valid filename
  ensureMapNameValidFilename();


  if( readMapInfo( archive ) < 0 )
  {
    return -1;
  }

  if( readt3HeightMap( archive ) < 0 )
  {
    return -1;
  }

  if( readt3CellFlags( archive ) < 0 )
  {
    return -1;
  }

  if( readObjects( archive ) < 0 )
  {
    return -1;
  }

  SFileCloseArchive( archive );
  return 0;
}



int SC2Map::readArchiveFile( const HANDLE archive,
                             const char*  strFilename,
                             int*         bufferOutSize,
                             u8**         bufferOut )
{
  HANDLE hFile;

  if( !SFileOpenFileEx( archive, strFilename, 0, &hFile ) )
  {
    printWarning( "Could not open %s for reading.\n", strFilename );
    return -1;
  }

  DWORD fileSizeBytes = SFileGetFileSize( hFile, NULL );
  if( fileSizeBytes == SFILE_INVALID_SIZE ||
      fileSizeBytes <= 0 )
  {
    printWarning( "%s is empty or invalid.\n", strFilename );
    return -1;
  }

  u8*   buffer = new u8[fileSizeBytes];
  DWORD numBytesRead;

  // initialize buffer to easily recognized values,
  // after a successful read all of them are overwritten
  memset( buffer, 'q', fileSizeBytes );

  if( !SFileReadFile( hFile, (void*)buffer, fileSizeBytes, &numBytesRead, NULL ) )
  {
    delete buffer;
    printWarning( "Could not read %s from archive.\n", strFilename );
    return -1;
  }

  if( numBytesRead != fileSizeBytes )
  {
    delete buffer;
    printWarning( "Could not read %s from archive. [NOT EXPECTING TO SEE THIS WARNING!]\n", strFilename );
    return -1;
  }

  *bufferOutSize = fileSizeBytes;
  *bufferOut     = buffer;

  SFileCloseFile( hFile );
  return 0;
}



int SC2Map::verifyMagicWord( const u8* magic, const u8* buffer )
{
  // a magic word (4 bytes) begins every file in an SC2Map
  // archive, to help verify its contents
  if( buffer[0] != magic[0] ||
      buffer[1] != magic[1] ||
      buffer[2] != magic[2] ||
      buffer[3] != magic[3] )
  {
    return -1;
  }

  return 0;
}



int SC2Map::readGameStrings_txt( const HANDLE archive )
{
  char strFilename[] = "enUS.SC2Data\\LocalizedData\\GameStrings.txt";

  int bufferSize;
  u8* bufferFreeAfterUse;

  if( readArchiveFile( archive, strFilename, &bufferSize, &bufferFreeAfterUse ) < 0 )
  {
    return -1;
  }

  // this is a plain text file, so copy the data into a null-terminated character
  // buffer so we can use string searching on it
  char bufferString[bufferSize + 1];
  memcpy( bufferString, bufferFreeAfterUse, bufferSize );
  bufferString[sizeof( bufferString ) - 1] = '\0';

  // release the buffer with the file's data
  delete bufferFreeAfterUse;

  // try to find the map's name here, otherwise use the
  // filename--the goal is to name the output files
  char attributeName[] = "DocInfo/Name=";

  char* findAttribute = strstr( bufferString, attributeName );

  if( findAttribute != NULL )
  {
    int nameStartIndex = (int)findAttribute -
                         (int)bufferString  +
                         strlen( attributeName );
    int i = 0;
    while( nameStartIndex + i < bufferSize &&
           bufferString[nameStartIndex + i] != '\r' )
    {
      this->strMapName[i] = bufferString[nameStartIndex + i];
      ++i;
    }
    strMapName[i] = '\0';

  } else {
    // otherwise return -1 because we should choose another name
    return -1;
  }

  return 0;
}


void SC2Map::setMapNameFromArchiveName( const char* strArchiveFilename )
{
  int i = 0;
  while( strArchiveFilename[i] != '.' )
  {
    this->strMapName[i] = strArchiveFilename[i];
    ++i;
  }
  strMapName[i] = '\0';
}


void SC2Map::ensureMapNameValidFilename()
{
  // strings we might find to name the map
  int fpos = 0;
  int npos = 0;

  if( !isalpha( strMapName[npos] ) )
  {
    char prefix[] = "sc2map";
    strcpy( strMapFilename, prefix );
    fpos = sizeof( prefix ) - 1;
  }

  while( strMapName[npos] != '\0' )
  {
    if( strMapName[npos] == ' ' ) {
      strMapFilename[fpos] = '_';
      ++fpos;

    } else if( strMapName[npos] == '.' ||
               strMapName[npos] == '-' ||
               strMapName[npos] == '(' ||
               strMapName[npos] == ')' ||
               isalnum( strMapName[npos] )
             ) {
      strMapFilename[fpos] = strMapName[npos];
      ++fpos;
    }

    ++npos;
  }

  strMapFilename[fpos] = '\0';

  // also throw a C++ version of the map FILENAME into a set of used map
  // filenames for this execution, so we can warn if one gets reused
  string mapFilename( strMapName );
  set<string>::iterator sItr = mapFilenamesUsed.find( mapFilename );
  if( sItr != mapFilenamesUsed.end() )
  {
    printWarning( "Map name %s begin reused, map-specific output for previous instance will be clobbered.\n", strMapFilename );
  }
  mapFilenamesUsed.insert( mapFilename );
}



int SC2Map::readMapInfo( const HANDLE archive )
{
  char strFilename[] = "MapInfo";

  int bufferSize;
  u8* bufferFreeAfterUse;

  if( readArchiveFile( archive, strFilename, &bufferSize, &bufferFreeAfterUse ) < 0 )
  {
    return -1;
  }

  bool continueProcessing = true;

  // before continuing, verify a minimum size in bytes
  // we are expecting until the variable-length strings
  int expectedSize =
    4+ // magic
    4+ // file version
    4+ // width
    4+ // height
    4+ // unknown
    4; // unknown

  // the file may still be invalid, but don't even
  // proceed if there isn't already this much--it
  // helps from having to check how many bytes are
  // left in all the following processing steps
  if( bufferSize < expectedSize )
  {
    printWarning( "%s is invalid.\n", strFilename );
    continueProcessing = false;
  }

  // position in the data buffer
  int buffpos = 0;

  if( continueProcessing )
  {
    // magic word to help verify file contents, always at file offset 0
    u8 magic[] = { 'I', 'p', 'a', 'M' };

    if( verifyMagicWord( magic, bufferFreeAfterUse ) < 0 )
    {
      printWarning( "Incorrect magic word for %s.\n", strFilename );
      continueProcessing = false;
    }

    buffpos += 4;
  }

  // raw data to read in
  u32 mcWidth;
  u32 mcHeight;
  u32 mcLeft;
  u32 mcBottom;
  u32 mcRight;
  u32 mcTop;

  if( continueProcessing )
  {
    // ignore file version (word)
    buffpos += 4;

    // width and height of map (word)
    memcpy( (void*) &mcWidth,  (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;
    memcpy( (void*) &mcHeight, (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;

    if( mcWidth > 256 || mcHeight > 256 )
    {
      printWarning( "Unexpected format for %s.\n", strFilename );
      continueProcessing = false;
    }
  }


  if( continueProcessing )
  {
    // ignore two unknown words
    buffpos += 8;

    // This section has been troublesome--sometimes a group of
    // zero bytes preceeds the first of two strings, and the
    // string sequence is as follows:

    // now there are two char strings of unknown length--starting
    // from the first string's offset look for a zero byte to
    // terminate the first string, and then a zero byte to terminate
    // the second string, and read more info from there

    // SO, first walk over any zero bytes before walking the strings
    while( buffpos < bufferSize && bufferFreeAfterUse[buffpos] == 0 )
    {
      ++buffpos;
    }

    // then walk the first string
    while( buffpos < bufferSize && bufferFreeAfterUse[buffpos] != 0 )
    {
      ++buffpos;
    }

    // advance past the termination character
    ++buffpos;

    // find terminator of second string
    while( buffpos < bufferSize && bufferFreeAfterUse[buffpos] != 0 )
    {
      ++buffpos;
    }

    // advance past the termination character
    ++buffpos;

    // there should be enough leftover data for 4 4-byte words
    if( bufferSize - buffpos < 16 )
    {
      printWarning( "Unexpected end-of-file in %s.\n", strFilename );
      continueProcessing = false;
    }
  }

  if( continueProcessing )
  {
    // use variable filepos to align the rest of the data
    memcpy( (void*) &mcLeft,   (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;
    memcpy( (void*) &mcBottom, (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;
    memcpy( (void*) &mcRight,  (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;
    memcpy( (void*) &mcTop,    (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;

    // there is a bunch more data in this file, but ignoring it

    if( mcLeft   > mcRight ||
        mcRight  > mcWidth ||
        mcBottom > mcTop   ||
        mcTop    > mcHeight )
    {
      printWarning( "Unexpected format for %s--likely an unknown, new version.\n", strFilename );
      continueProcessing = false;
    }
  }

  // whether or not we continued processing we have to
  // release the buffer with the file's data
  delete bufferFreeAfterUse;

  if( continueProcessing )
  {
    printMessage( "Map size in cells: %ux%u\n", mcWidth, mcHeight );
    printMessage( "Boundaries: (%u, %u) to  (%u, %u)\n", mcLeft, mcBottom, mcRight, mcTop );

    // the image frame constants are already defined,
    // (not read in by this function) but pass them
    // in now to set up translation so we can start
    // using points
    point::setFrameTranslationConstants( mcLeft,
                                         mcBottom,
                                         iDimT,
                                         iDimC );
    cxDimMap = mcWidth;
    cyDimMap = mcHeight;
    txDimMap = cxDimMap + 1;
    tyDimMap = cyDimMap + 1;

    cxDimPlayable = mcRight - mcLeft;
    cyDimPlayable = mcTop   - mcBottom;
    txDimPlayable = cxDimPlayable + 1;
    tyDimPlayable = cyDimPlayable + 1;

    // set the coordinates of the playable area in cell
    // coordinates, which they are natively given in
    cLeftBottom.mcSet( mcLeft,  mcBottom );
    cRightTop  .mcSet( mcRight, mcTop    );

    // this is a purposeful translation of cell coordinates
    // to terrain coordinates so we know the playable terrain
    tLeftBottom.mtSet( mcLeft,      mcBottom     );
    tRightTop  .mtSet( mcWidth + 1, mcHeight + 1 );

    // if we processed to the end with no problem,
    // report no problem
    return 0;
  }

  // otherwise report error
  return -1;
}



int SC2Map::readt3HeightMap( const HANDLE archive )
{
  char strFilename[] = "t3HeightMap";

  int bufferSize;
  u8* bufferFreeAfterUse;

  if( readArchiveFile( archive, strFilename, &bufferSize, &bufferFreeAfterUse ) < 0 )
  {
    return -1;
  }

  bool continueProcessing = true;

  // before continuing, verify a minimum size in bytes
  // we are expecting for this file
  int expectedSize =
    4+ // magic
    4+ // file version
    4+ // width
    4+ // height
    16;// unknown

  // the file may still be invalid, but don't even
  // proceed if there isn't already this much--it
  // helps from having to check how many bytes are
  // left in all the following processing steps
  if( bufferSize < expectedSize )
  {
    printWarning( "%s is invalid.\n", strFilename );
    continueProcessing = false;
  }

  // position in the data buffer
  int buffpos = 0;

  if( continueProcessing )
  {
    // magic word to help verify file contents, always at file offset 0
    u8 magic[] = { 'H', 'M', 'A', 'P' };

    if( verifyMagicWord( magic, bufferFreeAfterUse ) < 0 )
    {
      printWarning( "Incorrect magic word for %s.\n", strFilename );
      continueProcessing = false;
    }

    buffpos += 4;
  }

  // raw data to read in
  u32 mtWidth;
  u32 mtHeight;

  if( continueProcessing )
  {
    // ignore file version (word)
    buffpos += 4;

    // width and height of map (word)
    memcpy( (void*) &mtWidth,  (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;
    memcpy( (void*) &mtHeight, (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;

    if( mtWidth != txDimMap || mtHeight != tyDimMap )
    {
      // we are trying to read in the map of height data, and it should
      // match the dimensions we read in from MapInfo
      printWarning( "Terrain dimensions %dx%d in %s do not match %dx%d from MapInfo.\n",
                    mtWidth, mtHeight,
                    strFilename,
                    txDimMap, tyDimMap );
      continueProcessing = false;
    }

    // ignore 16 unknown bytes
    buffpos += 16;
  }

  if( continueProcessing )
  {
    // we are expecting the terrain map to have 6 bytes per terrain unit,
    // check that there is enough remaining data
    expectedSize = mtWidth*mtHeight*6;

    int bytesLeft = bufferSize - buffpos;

    if( bytesLeft < expectedSize )
    {
      printWarning( "Unexpected end-of-file in %s.\n", strFilename );
      continueProcessing = false;

    } else if( bytesLeft != bufferSize - buffpos ) {
      printWarning( "More data than expected in %s, continuing anyway.\n", strFilename );
    }
  }

  if( continueProcessing )
  {
    // the data for analysis only includes the
    // playable area, but...
    mapHeight = new u8[txDimPlayable*tyDimPlayable];

    // we gotta scan the whole file to get what we want
    for( int mtj = 0; mtj < tyDimMap; ++mtj )
    {
      for( int mti = 0; mti < txDimMap; ++mti )
      {
        // height is the 4th byte of each 6-byte chunk
        u8 h = bufferFreeAfterUse[buffpos + 4]; buffpos += 6;

        if( h > 0x3 )
        {
          // only 0x0, 0x1, 0x2 and 0x3 are valid
          printWarning( "Unexpected value %u in %s\n", h, strFilename );
          continueProcessing = false;
          break;
        }

        point t;
        t.mtSet( mti, mtj );
        if( isPlayableTerrain( &t ) )
        {
          setHeight( &t, h );
        }
      }

      if( !continueProcessing )
      {
        break;
      }
    }
  }

  // whether or not we continued processing we have to
  // release the buffer with the file's data
  delete bufferFreeAfterUse;

  if( continueProcessing )
  {
    // if we processed to the end with no problem,
    // report no problem
    return 0;
  }

  // otherwise report error
  return -1;
}


void SC2Map::setHeight( point* t, u8 h )
{
  if( !isPlayableTerrain( t ) )
  {
    printError( "Attempt to access data from unplayable terrain (%d, %d)\n",
                t->ptx,
                t->pty );
    exit( -1 );
  }
  mapHeight[t->pty*txDimPlayable + t->ptx] = h;
}


u8 SC2Map::getHeight( point* t )
{
  if( !isPlayableTerrain( t ) )
  {
    printError( "Attempt to access data from unplayable terrain (%d, %d)\n",
                t->ptx,
                t->pty );
    exit( -1 );
  }
  return mapHeight[t->pty*txDimPlayable + t->ptx];
}



int SC2Map::readt3CellFlags( const HANDLE archive )
{
  char strFilename[] = "t3CellFlags";

  int bufferSize;
  u8* bufferFreeAfterUse;

  if( readArchiveFile( archive, strFilename, &bufferSize, &bufferFreeAfterUse ) < 0 )
  {
    return -1;
  }

  bool continueProcessing = true;

  // before continuing, verify a minimum size in bytes
  // we are expecting for this file
  int expectedSize =
    4+ // magic
    4+ // file version
    16+// unknown
    4+ // width
    4; // height

  // the file may still be invalid, but don't even
  // proceed if there isn't already this much--it
  // helps from having to check how many bytes are
  // left in all the following processing steps
  if( bufferSize < expectedSize )
  {
    printWarning( "%s is invalid.\n", strFilename );
    continueProcessing = false;
  }

  int buffpos = 0;

  if( continueProcessing )
  {
    // magic word to help verify file contents, always at file offset 0
    u8 magic[] = { 'L', 'F', 'C', 'T' };

    if( verifyMagicWord( magic, bufferFreeAfterUse ) < 0 )
    {
      printWarning( "Incorrect magic word for %s.\n", strFilename );
      continueProcessing = false;
    }

    buffpos += 4;
  }

  // raw data to read in
  u32 mcWidth;
  u32 mcHeight;

  if( continueProcessing )
  {
    // ignore file version (word) and an unknown 16 bytes
    buffpos += 20;

    // width and height of map (word)
    memcpy( (void*) &mcWidth,  (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;
    memcpy( (void*) &mcHeight, (void*) &(bufferFreeAfterUse[buffpos]), 4 ); buffpos += 4;

    if( mcWidth != cxDimMap || mcHeight != cyDimMap )
    {
      // we are trying to read in the map of height data, and it should
      // match the dimensions we read in from MapInfo
      printWarning( "Cell dimensions %dx%d in %s do not match %dx%d from MapInfo.\n",
                    mcWidth, mcHeight,
                    strFilename,
                    cxDimMap, cyDimMap );
      continueProcessing = false;
    }
  }

  if( continueProcessing )
  {
    // we are expecting the cell map to have 1 byte per cell,
    // check that there is enough remaining data
    expectedSize = mcWidth*mcHeight;

    int bytesLeft = bufferSize - buffpos;

    if( bytesLeft < expectedSize )
    {
      printWarning( "Unexpected end-of-file in %s.\n", strFilename );
      continueProcessing = false;

    } else if( bytesLeft != bufferSize - buffpos ) {
      printWarning( "More data than expected in %s, continuing anyway.\n", strFilename );
    }
  }

  if( continueProcessing )
  {
    numPathableCells[PATH_GROUND_NOROCKS]   = 0;
    numPathableCells[PATH_GROUND_WITHROCKS] = 0;
    numPathableCells[PATH_CWALK_NOROCKS]    = 0;
    numPathableCells[PATH_CWALK_WITHROCKS]  = 0;


    // the data for analysis only includes the
    // playable area, but...
    mapPathing = new bool[cxDimPlayable*cyDimPlayable*NUM_PATH_TYPES];

    // we gotta scan the whole file to get what we want
    for( int mcj = 0; mcj < cyDimMap; ++mcj )
    {
      for( int mci = 0; mci < cxDimMap; ++mci )
      {
        u8 flags = bufferFreeAfterUse[buffpos]; buffpos += 1;

        point c;
        c.mcSet( mci, mcj );
        if( isPlayableCell( &c ) )
        {
          // take the 2nd LSB as indicator for not-pathable
          // so flip flags first to get "pathable by ground"
          bool groundPathable = ~flags & 0x1;

          // if any of the bordering terrain units are the
          // lowest elevation then make this cell unpathable
          bool everPathable = true;
          point tlb, trb, tlt, trt;

          tlb.ptSet( c.pcx,     c.pcy     );
          trb.ptSet( c.pcx + 1, c.pcy     );
          tlt.ptSet( c.pcx,     c.pcy + 1 );
          trt.ptSet( c.pcx + 1, c.pcy + 1 );

          u8 hlb = getHeight( &tlb );
          u8 hrb = getHeight( &trb );
          u8 hlt = getHeight( &tlt );
          u8 hrt = getHeight( &trt );

          if( hlb == 0x0 ||
              hrb == 0x0 ||
              hlt == 0x0 ||
              hrt == 0x0 )
          {
            everPathable = false;
          }

          // also, for any cell with bordering terrain units
          // with a difference in elevation > 1 we also no
          // the cell is never pathable (cliff-walkers)
          if( abs( hlb - hrb ) > 1 ||
              abs( hlb - hlt ) > 1 ||
              abs( hlb - hrt ) > 1 ||
              abs( hrb - hlt ) > 1 ||
              abs( hrb - hrt ) > 1 ||
              abs( hlt - hrt ) > 1 )
          {
            everPathable = false;
          }

          // set the pathing for all path types--the paths that
          // consider destructible rocks should be set here and
          // will be modified later when reading the Objects file
          setPathing( &c, PATH_GROUND_NOROCKS,   false );
          setPathing( &c, PATH_GROUND_WITHROCKS, false );
          setPathing( &c, PATH_CWALK_NOROCKS,    false );
          setPathing( &c, PATH_CWALK_WITHROCKS,  false );

          if( everPathable )
          {
            if( groundPathable )
            {
              numPathableCells[PATH_GROUND_NOROCKS]++;
              numPathableCells[PATH_GROUND_WITHROCKS]++;
              setPathing( &c, PATH_GROUND_NOROCKS,   true );
              setPathing( &c, PATH_GROUND_WITHROCKS, true );
            }

            numPathableCells[PATH_CWALK_NOROCKS]++;
            numPathableCells[PATH_CWALK_WITHROCKS]++;
            setPathing( &c, PATH_CWALK_NOROCKS,   true );
            setPathing( &c, PATH_CWALK_WITHROCKS, true );
          }
        }
      }
    }
  }


  // whether or not we continued processing we have to
  // release the buffer with the file's data
  delete bufferFreeAfterUse;

  if( continueProcessing )
  {
    // if we processed to the end with no problem,
    // report no problem
    return 0;
  }

  // otherwise report error
  return -1;
}


void SC2Map::setPathing( point* c, PathType t, bool p )
{
  if( !isPlayableCell( c ) )
  {
    printError( "Attempt to access data from unplayable cell (%d, %d).\n",
                c->pcx,
                c->pcy );
    exit( -1 );
  }
  mapPathing[NUM_PATH_TYPES*(c->pcy*cxDimPlayable + c->pcx) + t] = p;
}


bool SC2Map::getPathing( point* c, PathType t )
{
  if( !isPlayableCell( c ) )
  {
    printError( "Attempt to access data from unplayable cell (%d, %d)\n.",
                c->pcx,
                c->pcy );
    exit( -1 );
  }
  return mapPathing[NUM_PATH_TYPES*(c->pcy*cxDimPlayable + c->pcx) + t];
}



int SC2Map::readObjects( const HANDLE archive )
{
  char strFilename[] = "Objects";

  int bufferSize;
  u8* bufferFreeAfterUse;

  if( readArchiveFile( archive, strFilename, &bufferSize, &bufferFreeAfterUse ) < 0 )
  {
    return -1;
  }

  bool continueProcessing = true;

  // we need a FILE* instead of data buffer for tinyxml lib
  char strTemp[] = "Objects.tmp";
  FILE* fileTemp = fopen( strTemp, "w+" );

  if( fileTemp == NULL )
  {
    printError( "Could not open temp file for %s.\n", strFilename );
    continueProcessing = false;
  }

  if( continueProcessing )
  {
    int bytesWritten = fwrite( (char*)bufferFreeAfterUse, 1, bufferSize, fileTemp );
    if( bytesWritten != bufferSize )
    {
      printError( "Could not dump %s to temp file.\n", strFilename );
      continueProcessing = false;
    }

    // from here on out a temporary file holds the XML
    // so after processing delete the file
    fclose( fileTemp );
  }

  // whether or not we continue processing we have to
  // release the buffer with the file's data
  delete bufferFreeAfterUse;


  bool continueXML = false;
  if( continueProcessing )
  {
    continueXML = true;
  }

  TiXmlDocument doc;

  if( continueXML )
  {
    // open temp file as XML stream
    if( !doc.LoadFile( strTemp ) )
    {
      printError( "Could not open temp file for %s.\n", strFilename );
      continueXML = false;
    }
  }

  TiXmlHandle docHandle( &doc );
  TiXmlElement* placedObjects = NULL;

  if( continueXML )
  {
    placedObjects = docHandle
      .FirstChildElement( "PlacedObjects" )
      .ToElement();
    if( !placedObjects )
    {
      printWarning( "%s is not valid.\n", strFilename );
      continueXML = false;
    }
  }

  if( continueXML )
  {
    const char* strVersion = placedObjects->Attribute( "version" );
    if( strVersion == NULL )
    {
      printWarning( "%s has an unknown version.\n", strFilename );
      continueXML = false;
    }
    else if( strcmp( strVersion, "26" ) != 0 )
    {
      printWarning( "%s is version %s, file version 26 expected.\n", strFilename, strVersion );
    }
  }

  if( continueXML )
  {
    for( TiXmlElement* object = placedObjects->FirstChildElement();
         object;
         object = object->NextSiblingElement() )
    {
      // in placedobjects.cpp
      processPlacedObject( object );
    }
  }

  // delete the temporary file
  char cmd[512];
  sprintf( cmd, "if exist %s del %s", strTemp, strTemp );
  system( cmd );

  if( continueProcessing && continueXML )
  {
    return 0;
  }

  return -1;
}
